#pragma once

#include "types.h"
#include "pack_export.h"


//class PackBase : public PackSetup
//class __declspec(dllexport) PackBase : public PackSetup
class PackBase : public PackSetup, public PackExports
{
#if 1
// stubs
public:
	bool SetHdrBuf ();
	bool HdrParse (bool& is_valid);
	bool TypeSetup ();
#endif
private:
// prevent an instantiation of this class (and above, until the top point)
// note that this will produce vtab (add 8 bytes at the size of this class), which will be erased by zeroing at mem alloc time, but we don't use vtab here anyway
	virtual void Dummy () = 0;
	void ConvertError ();
public:
//	PackBase ();
	bool Open (bool write, bool widechar, const void* filepath);
	bool Close ();
	bool Delete ();
	bool SetHandlerByName (const char* name);
// *'ReadHeader()' conflicts at linking
	bool ReadHdr ();
	bool PrepareHdr (u32 files_num);
// *no even 'bool swap_filepointers' (which is mandatory), because it only adds a mess
	bool SetupContainer ();
	bool HasMoreFiles (bool& has_more);
	bool ReadEntry ();
	bool GetEntryNameRequiredBufSize (u32& buf_size);
	bool GetEntryName (char* buf);
private:
	bool WriteEntry ();
public:
	bool SkipFileData ();
private:
#if 0
	bool ReadToMem ();
	bool WriteFromMem ();
#endif
	bool CopyToFile_Mallocated (bool widechar, const void* filepath);
public:
	bool CopyToFile (bool widechar, const void* filepath);
private:
	bool CheckFileNameForEntry ();
	bool PutFileName (bool widechar, const void* filename);
public:
//	bool CopyFromFile (bool widechar, const void* filepath);
	bool CopyFromFile (bool widechar, const void* path, const void* filename);
	bool FinishWrite ();

// *note, that this is too much of stuff here :  such a number of pointers perhaps doesn't make the code legible, and it would maybe be better for a pack handlers themselves to read/write these values from their formats
// currently :  reading is done by us, writing is done by a handler
// or even may appear, that this improportionality is the best way
// yet the name is being copied by us, because it may need a conversion from widechar
// *now the code must be more legible :  all stuff which is being set up, was moved to 'PackSetup'

protected:
//private:
//	bool m_is_opened;
	File m_file;
private:
	void* m_mem;
// *helper
//	bool m_write;
public:
// *note, that I set it to -1, when the operation is not yet completed/finished (some error still may occur ahead; useable in a private functions)
//	u32 m_error;
#if 0
// remember, what name it has
	char m_arc_name[1024];
//	WCHAR m_arc_name[1024];
#endif
// filetime of this pack
	FILETIME m_packfile_filetime;
// filetime to use for a file being unpacked
	FILETIME m_filetime;
//	char m_name[HOG_ENTRY_FILE_NAME_BUF_SIZE];
// length of data processed or to process
	u32 m_data_len;
// storage for recognition by tag
//	char m_file_beginning[32];
	char m_buf_for_tag[32];
//// buffer for conversion from widechar
//	char m_buf_for_name[1024];
#if 0
// +++ list of containers +++
	pack_handler_t* m_handlers_array;
	u32 m_handlers_num;
// --- list of containers ---
#endif
// +++ of single block +++
// hmmm, this var is superfluous; use 'm_file.m_pos' instead
//	u32 m_compound_offset_cur;
//	u32 m_common_offset_cur;
// --- of single block ---
// +++ of two blocks +++
// *files number may appear in a various smaller containers (u24, u16, u8), so I am not making a pointer to it - the header parser routine should read it
	u32 m_files_num;
	u32 m_entries_begin;
	u32 m_entry_offset_cur;
	u32 m_data_begin;
	u32 m_data_offset_cur;
// *if header has this; otherwise it is 0
	u32 m_file_end;
// --- of two blocks ---
// +++ current file +++
	u32 m_file_ix;
// the content of entry itself
	bool m_is_entry_good;
// availability of data which is pointed by the entry (it is an error otherwise)
	bool m_has_entry_data;
// --- current file ---
// +++ indication +++
// *actually could be used for 'Open()' also
	bool m_info_filename_is_widechar;
	const void* m_info_filename;
//	bool m_info_filepath_is_widechar;
//	const void* m_info_filepath;
// *you may reset it to 0 whenever you need
	u32 m_data_bytes_processed;
// --- indication ---
};

#define FRD(_obj) m_file.Read (&_obj, sizeof(_obj))
#define FWR(_obj) m_file.Write (&_obj, sizeof(_obj))
// with error check
#define FRD_WEC(_obj) \
do \
{ \
	r = FRD(_obj); \
	if (! r) \
		goto file_error; \
} while (0)

#define FWR_WEC(_obj) \
do \
{ \
	r = FWR(_obj); \
	if (! r) \
		goto file_error; \
} while (0)

