// by Alexander Ilyin
// 'PackTop' is the top of the rhomb structure above the 'Pack' classes
// once need to add one more packer, add its class at 'class PackTop', and then its properties and functions into '___' array of 'pack_handler_t'
// this trick is necessary because can't define a function outside of 'Pack', without having declared one in it

// *'Boost::enable_if' has been introduced in Boost 1.30.0, and in a separate file "boost/utility/enable_if.hpp" it been made in Boost 1.31.0 (after revision at 2003-12-02)
// but it didn't support MSVC 7.0 and lower

#include "pack_top.h"
#include "macros.h"


// *MSVCs above 6.0 do have '__if_exists()'/'__if_not_exists()' (it is a bit more than SFINAE)
#define FUNC_IF_PACK_OVERLOAD_EXISTS(_class, _func) (&_class::_func == &PackBase::_func ? 0 : &_class::_func)
#define HDR_FUNCS_IF_PACK_OVERLOADS_EXIST(_class) \
FUNC_IF_PACK_OVERLOAD_EXISTS(_class, SetHdrBuf), \
FUNC_IF_PACK_OVERLOAD_EXISTS(_class, HdrParse), \
FUNC_IF_PACK_OVERLOAD_EXISTS(_class, TypeSetup),

pack_handler_t pack_handlers[] =
{
#if 0
//	{"DHF", 0, 0, 0, &Hog::TypeSetup},
//	{"HOG2", &Hog2::m_buf_hdr, sizeof(Hog2::m_buf_hdr), 0, &Hog2::TypeSetup},
//	{"HOG2", 0, 0, 0, &Hog2::TypeSetup},
	{"hog", "DHF", 0, 0, &Hog::TypeSetup},
	{"hog2", "HOG2", &Hog2::SetHdrBuf, &Hog2::HdrParse, &Hog2::TypeSetup},
	{"pak", "PACK", &Pak::SetHdrBuf, &Pak::HdrParse, &Pak::TypeSetup},
#endif
#if 1
	{"hog", "hog", "DHF", HDR_FUNCS_IF_PACK_OVERLOADS_EXIST(Hog)},
	{"hog2", "hog", "HOG2", HDR_FUNCS_IF_PACK_OVERLOADS_EXIST(Hog2)},
	{"pak", "pak", "PACK", HDR_FUNCS_IF_PACK_OVERLOADS_EXIST(Pak)},
	{"iwad", "wad", "IWAD", HDR_FUNCS_IF_PACK_OVERLOADS_EXIST(Wad)},
	{"pwad", "wad", "PWAD", HDR_FUNCS_IF_PACK_OVERLOADS_EXIST(Wad)},
	{"wad2", "wad", "WAD2", HDR_FUNCS_IF_PACK_OVERLOADS_EXIST(Wad3)},
	{"wad3", "wad", "WAD3", HDR_FUNCS_IF_PACK_OVERLOADS_EXIST(Wad3)},
#endif
// or this way
#if 0
	{"hog", "DHF", CLASS_HEADER_FUNCS(Hog)},
	{"hog2", "HOG2", CLASS_HEADER_FUNCS(Hog2)},
	{"pak", "PACK", CLASS_HEADER_FUNCS(Pak)},
#endif
//	0
};

//u32 pack_handlers_num = NUMOF(pack_handlers);
const u32 pack_handlers_num = NUMOF(pack_handlers);

const char* pack_format_names[pack_handlers_num] = {0};
//const char* pack_format_names[1] = {0};


PackTop::PackTop ()
// *it calls 'PackBase ()' anyway
//:
//PackTop ()
//PackBase ()
{
// clear the whole 'Pack', but don't clear our vtabs, because it will not be nice :)
// *zeroing in the default constructor was introduced in C++-03
//	memset ((PackBase*)this, 0, sizeof(PackBase));
// subtract the place for pointers
//	memset (&this->_m_beg, 0, sizeof(PackBase) - (&this->_m_beg - (u8*)(PackBase*)this));
// subtract the place for pointers and '_m_beg' marker
	memset (&this->_m_beg + 1, 0, sizeof(PackBase) - (&this->_m_beg - (u8*)(PackBase*)this) - 1);
// 'Pack' will have to call functions of a particular format classes
	m_pack_top = this;
}

void PackTop::GetFormatNames (const char**& format_names_list, u32& names_num)
{
// fill the list
	FORI0(pack_handlers_num)
	{
		pack_format_names[i] = pack_handlers[i].name;
	}
	format_names_list = pack_format_names;
	names_num = pack_handlers_num;
}

