// by Alexander Ilyin
// Total/Double Commander plugin

#include "plugin.h"
#include "error.h"
#include "pack_top.h"
#include <string>
using namespace std;
#include "cfg_dlg.h"
#include "macros.h"


HINSTANCE g_hinstance = 0;
// +++ settings +++
//bool is_hog2 = 0;
u32 packtype_hog = 0;
u32 packtype_wad = 1;
// --- settings ---
// *hmmm, it shouldn't be global, but it must belong at appropriate 'Pack'
tProcessDataProc g_process_data_proc = 0;
tProcessDataProcW g_process_data_proc_w = 0;


#if 0
template<typename T>
T* new0 ()
{
	T* obj;
	obj = new T;
	if (obj)
		memset (obj, 0, sizeof(*obj));
	return obj;
}
#endif
#if 0
template<typename T>
void new0 (T*& obj)
{
	obj = new T;
	if (obj)
		memset (obj, 0, sizeof(*obj));
}
#endif

BOOL APIENTRY DllMain (HINSTANCE hinst, DWORD reason, LPVOID reserved)
{
	switch (reason)
	{
	case DLL_PROCESS_ATTACH:
		g_hinstance = hinst;
		break;
	case DLL_PROCESS_DETACH:
		break;
	default:
		break;
	}
	return 1;
}

// *note, that this makes an incorrect transformation; for example 'ERROR_SHARING_VIOLATION' will produce 'FILE_ERR__NO_ACCESS' at either open or write,
// and here it will be transformed into 'E_EOPEN', but open and write errors is a different thing
u32 ConvertError (PackTop* pack)
{
	A(pack);
	switch (pack->m_error)
	{
	case PACK_ERR__NONE:
//		return 0;
// this function should not be called if there was no error
// (theoretically it could be called in any case, but for now I made it to be called only in the case of an error)
		ET;
	case PACK_ERR__OUT_OF_MEMORY:
		return E_NO_MEMORY;
	case PACK_ERR__UNSPECIFIED_OPEN_FOR_READ_ERROR:
		return E_EOPEN;
	case PACK_ERR__UNSPECIFIED_OPEN_FOR_WRITE_ERROR:
		return E_EWRITE;
	case PACK_ERR__NO_FILE:
	case PACK_ERR__NO_ACCESS:
		return E_EOPEN;
	case PACK_ERR__CLOSE_ERROR:
		return E_ECLOSE;
	case PACK_ERR__UNSPECIFIED_DELETE_ERROR:
		return E_ECLOSE;
	case PACK_ERR__READ_FAULT:
		return E_EREAD;
// *would be good to report about "disk full", but Total Commander doesn't offer that error
	case PACK_ERR__WRITE_FAULT:
	case PACK_ERR__DISK_FULL:
		return E_EWRITE;
	case PACK_ERR__UNKNOWN_FORMAT:
//		return E_BAD_ARCHIVE;
	case PACK_ERR__UNSUPPORTED_FORMAT_VERSION:
		return E_UNKNOWN_FORMAT;
	case PACK_ERR__BAD_ENTRY:
	case PACK_ERR__BAD_FILE:
		return E_BAD_ARCHIVE;
	case PACK_ERR__TOO_MUCH_ENTRIES:
		return E_BAD_ARCHIVE;
	case PACK_ERR__FILE_TOO_BIG:
		return E_NOT_SUPPORTED;
	case PACK_ERR__FILENAME_INCOMPATIBLE:
		return E_NOT_SUPPORTED;
	case PACK_ERR__UNSUPPORTED_ENTRY_TYPE:
		return E_NOT_SUPPORTED;
// deal with it before going here
	case PACK_ERR__SPECIAL_ENTRY:
		ET;
	default:
		ET;
	}
	return 0;
}

// the sequence of calls is :
// OpenArchive, SetProcessDataProc, SetChangeVolProc, {ReadHeader, ProcessFile}, CloseArchive

//// hmmm, can't do such a dispatcher
//static HANDLE OpenArch (* archive_data)
static HANDLE OpenArch (int& open_result, bool widechar, void* filepath)
{
	bool r;
	PackTop* pack;

	A(filepath);

//	new0 (pack);
	pack = new PackTop;
//	memset ((Pack*)pack, 0, sizeof(Pack));
// no, this is too complex for a compiler, to make a deduce by return type
// { A* a; a = t (); }   struct A {};   template<typename T> T* t () { return 0; }
	if (! pack)
	{
		open_result = E_NO_MEMORY;
		return 0;
	}

// yes, this is the mostly appropriate place for zeroing, but ... :/
#if 0
// *members are not zeroed by the default constructor
// zero the values
// *it may look that this initialization must be done inside of the classes, but that isn't so correct;
// for example :  what about the included 'm_file' ? - should it be zeroed inside of that class, or in the class which contains it ?
// '*obj = {0}' will not work anyway, since there are a private members
//	*pack = {0};
	memset (pack, 0, sizeof(*pack));
#endif
//	pack->m_handlers_array = pack_handlers;
//	pack->m_handlers_num = pack_handlers_num;

// *'archive_data->OpenMode' doesn't matter, because has to open the file for read anyway
	r = pack->Open (0, widechar, filepath);
	if (! r)
	{
//// hm, different from what I expected, it doesn't want to know about 'm_err_no_file' and 'm_err_no_access'
//		open_result = E_EOPEN;
		open_result = ConvertError (pack);
		goto free_obj;
	}

	r = pack->ReadHdr ();
	if (! r)
	{
		open_result = ConvertError (pack);
		r = pack->Close ();
		goto free_obj;
	}

	A(! pack->m_error);

	return pack;

free_obj:
	delete pack;
	return 0;
}

// *further 'ReadHeader()' will be called, if from here return 0 error in 'archive_data->OpenResult'
// and if not implement the ANSI functions, then this plugin will not work on older commanders
HANDLE __stdcall OpenArchive (tOpenArchiveData* archive_data)
{
	A(archive_data);
	return OpenArch (archive_data->OpenResult, 0, archive_data->ArcName);
}

// *widechar version of these functions requires Total Commander 7.5
HANDLE __stdcall OpenArchiveW (tOpenArchiveDataW* archive_data)
{
	A(archive_data);
	return OpenArch (archive_data->OpenResult, 1, archive_data->ArcName);
}

int __stdcall CloseArchive (HANDLE h_arc_data)
{
	bool r;
	PackTop* pack;

	A(h_arc_data);
	pack = (PackTop*)h_arc_data;
	r = pack->Close ();
	delete pack;
	if (! r)
	{
		return E_ECLOSE;
	}
	g_process_data_proc_w = 0;
	g_process_data_proc = 0;
	return 0;
}

// *for this HOG project, it is more convenient to use ANSI version of 'ReadHeader()'
int __stdcall ReadHeader (HANDLE h_arc_data, tHeaderData* header_data)
{
	bool r;
	PackTop* pack;
	bool has_more;
	u32 i;

#if 1
	A(h_arc_data);
#else
	if (! h_arc_data)
//		return 0;
		return E_BAD_DATA;
#endif
	A(header_data);
	pack = (PackTop*)h_arc_data;
//	A(pack->);
retry:
	r = pack->HasMoreFiles (has_more);
	if (! r)
	{
		i = ConvertError (pack);
		return i;
//		return E_EREAD;
	}
	if (! has_more)
	{
		return E_END_ARCHIVE;
	}

	r = pack->ReadEntry ();
	if (! r)
	{
// special case
		if (pack->m_error == PACK_ERR__SPECIAL_ENTRY)
// skip it
			goto retry;
		i = ConvertError (pack);
		return i;
	}

	header_data->PackSize = pack->m_entry_len + pack->m_data_len;
	header_data->UnpSize = pack->m_data_len;
// *this time isn't stored in the pack, so have to replace it with something
//	header_data->FileTime = pack->m_packfile_filetime;
	{
		FILETIME ft;
		SYSTEMTIME st;
//		i = FileTimeToLocalFileTime (&pack->m_packfile_filetime, &ft);
		i = FileTimeToLocalFileTime (&pack->m_filetime, &ft);
		i = FileTimeToSystemTime (&ft, &st);
		header_data->FileTime = (st.wYear - 1980) << 25 | st.wMonth << 21 | st.wDay << 16 | st.wHour << 11 | st.wMinute << 5 | st.wSecond / 2;
	}
// we must have space for that name and ending 0
#if 0
// *we currently don't have a formats which could require a higher size
	A(sizeof(header_data->FileName) > pack->m_name_len_max);
	memcpy (header_data->FileName, pack->m_name, pack->m_name_len_max);
//	memcpy (header_data->FileName, pack->m_name, pack->m_name_len_max + 1);
//// *even though it is stored in a HOG file as 0, it still should ensured to be 0, instead of just being copied from there
	header_data->FileName[pack->m_name_len_max] = 0;
#else
	u32 buf_size;
	r = pack->GetEntryNameRequiredBufSize (buf_size);
	A(r);
// *we currently don't have a formats which could require a higher size
	A(buf_size <= sizeof(header_data->FileName));

	r = pack->GetEntryName (header_data->FileName);
	A(r);
#endif

// is this necessary ?  for what ?
//	strcpy (header_data->ArcName, pack->m_arc_name);

	A(! pack->m_error);

	return 0;
}

// show progress
// requires the processed length and file name
bool ShowProgress (int& result, PackTop* pack)
{
	u32 i;
	const void* filename;

	result = -1;
	filename = pack->m_info_filename;
// try first to output it with the filename, but don't do a conversion, but try to call an appropriate function (let the conversion be done outside)
	if (pack->m_info_filename_is_widechar)
	{
		if (g_process_data_proc_w)
			goto output_w;
	}
	else
	{
		if (g_process_data_proc)
			goto output;
	}

// this we don't expect to occur, but still, try to output the progress, but without the filename
	filename = 0;
	if (g_process_data_proc_w)
	{
output_w:
//		i = g_process_data_proc_w (0, pack->m_data_bytes_processed);
		i = g_process_data_proc_w ((wchar_t*)filename, pack->m_data_bytes_processed);
	}
	else if (g_process_data_proc)
	{
output:
// *by "wcx_xz10_src" I see that it is the decompressed size
//		i = g_process_data_proc (0, pack->m_data_bytes_processed);
		i = g_process_data_proc ((char*)filename, pack->m_data_bytes_processed);
	}
	else
// the call of this subroutine isn't so necessary, but the lack of it is abnormal
	{
		ET;
	}

//finish:
	pack->m_data_bytes_processed = 0;
	if (! i)
	{
		pack->m_error = PACK_ERR__ABORTED;
		result = E_EABORTED;
		return 0;
	}

	result = 0;
	return 1;
}

//static bool ProcessFile_ (HANDLE h_arc_data, int operation, bool widechar, void* filepath)
static int ProcessFile_ (HANDLE h_arc_data, int operation, bool widechar, void* filepath)
{
	bool r;
	int result;
	PackTop* pack;
//	u32 i;

	A(h_arc_data);
	result = -1;
	pack = (PackTop*)h_arc_data;
	switch (operation)
	{
	case PK_SKIP:
	case PK_TEST:
		r = pack->SkipFileData ();
		if (! r)
			goto convert_error;
		break;
	case PK_EXTRACT:
		r = pack->CopyToFile (widechar, filepath);
		if (! r)
			goto convert_error;
// *filename of the output
		pack->m_info_filename_is_widechar = widechar;
		pack->m_info_filename = filepath;
		r = ShowProgress (result, pack);
		if (! r)
			goto error;
		break;
	}

	A(! pack->m_error);
//	result = 0;

//finish:
	return 0;
//	return result;

convert_error:
	result = ConvertError (pack);
error:
	A(result);
	return result;
}

// *XZ plugin is by Ghisler, and 'dest_path' argument in it is just ignored
int __stdcall ProcessFile (HANDLE h_arc_data, int operation, LPSTR dest_path, LPSTR dest_name)
{
	return ProcessFile_ (h_arc_data, operation, 0, dest_name);
}

int __stdcall ProcessFileW (HANDLE h_arc_data, int operation, wchar_t* dest_path, wchar_t* dest_name)
{
	return ProcessFile_ (h_arc_data, operation, 1, dest_name);
}

// note, that Total Commander calls this function only at the moment when a plugin is being installed, but later it is not being called anymore
// thus you may not get some expected behaviour, after trying to change this
// and also even in Total Commander 8.01 it isn't easy to remove a registered plugin from the list of plugins; to do that, open "WINCMD.INI" (which may be in a hidden folder), and remove the line with your plugin in [PackerPlugins] section
// otherwise you also may remove the number in that string and comma after it; this will make Total Commander to call this function again at its next start
int __stdcall GetPackerCaps ()
{
// we don't support repacking, thus PK_CAPS_MODIFY and PK_CAPS_DELETE are off
// *PK_CAPS_SEARCHTEXT is necessary for compare directories by content (otherwise there will be red question marks as the result)
	return PK_CAPS_NEW | PK_CAPS_MULTIPLE | PK_CAPS_OPTIONS | PK_CAPS_BY_CONTENT | PK_CAPS_SEARCHTEXT;
//	return PK_CAPS_NEW | PK_CAPS_MODIFY | PK_CAPS_MULTIPLE | PK_CAPS_DELETE | PK_CAPS_OPTIONS | PK_CAPS_MEMPACK	| PK_CAPS_BY_CONTENT | PK_CAPS_SEARCHTEXT;
}

void __stdcall SetChangeVolProc (HANDLE h_arc_data, tChangeVolProc change_vol_proc)
{
	A(h_arc_data);
}

void __stdcall SetChangeVolProcW (HANDLE h_arc_data, tChangeVolProcW change_vol_proc)
{
	A(h_arc_data);
}

void __stdcall SetProcessDataProc (HANDLE h_arc_data, tProcessDataProc process_data_proc)
{
	A(h_arc_data);
	A(! g_process_data_proc);
	g_process_data_proc = process_data_proc;
}

void __stdcall SetProcessDataProcW (HANDLE h_arc_data, tProcessDataProcW process_data_proc)
{
	A(h_arc_data);
	A(! g_process_data_proc_w);
	g_process_data_proc_w = process_data_proc;
}

bool CanYouHandleThisFile_Opened (PackTop* pack)
{
	bool r;
	bool has_more;

// file of 0 size isn't valid
	r = pack->ReadHdr ();
	if (! r)
	{
		return 0;
	}

//// if the remained file size is 0, then it is ok (no files inside of it)
	r = pack->HasMoreFiles (has_more);
	if (! r)
	{
		return 0;
	}
// no files inside of it -  it is allowed
	if (! has_more)
		return 1;

// try one entry
	r = pack->ReadEntry ();
	if (! r)
	{
		return 0;
	}

	return 1;
}

bool CanYouHandleThisFile_Pack (PackTop* pack, bool widechar, void* filepath)
{
	bool r;
	bool result;

	r = pack->Open (0, widechar, filepath);
	if (! r)
	{
		return 0;
	}

	result = CanYouHandleThisFile_Opened (pack);

	r = pack->Close ();
	if (! r)
	{
		return 0;
	}

	return result;
}

// *for Ctrl+PgDn
static bool CanYouHandleThisFile (bool widechar, void* filepath)
{
	bool r;
//	bool result;
	PackTop* pack;

	A(filepath);

//	new0 (pack);
	pack = new PackTop;
	if (! pack)
	{
		return 0;
	}

	r = CanYouHandleThisFile_Pack (pack, widechar, filepath);

	delete pack;

	return r;
}

BOOL __stdcall CanYouHandleThisFile (LPSTR filepath)
{
	return CanYouHandleThisFile (0, filepath);
}

BOOL __stdcall CanYouHandleThisFileW (wchar_t* filepath)
{
	return CanYouHandleThisFile (1, filepath);
}

void __stdcall ConfigurePacker (HWND parent, HINSTANCE dll_instance)
{
	BOOL B_r;
	int i;

	i = CfgDlg (parent, dll_instance);
	if (i != IDOK)
		return;
	i = i;
#define PLUGIN_SECTION_NAME "packformat"
#define PLUGIN_INI_NAME "wcx_pack.ini"
	char value[2] = {0};
	value[0] = '0' + packtype_hog;
	B_r = WritePrivateProfileString (PLUGIN_SECTION_NAME, "hog", value, PLUGIN_INI_NAME);
	value[0] = '0' + packtype_wad;
	B_r = WritePrivateProfileString (PLUGIN_SECTION_NAME, "wad", value, PLUGIN_INI_NAME);
}

void ReadConfigStringToU32 (u32& value, const char* section_name, const char* value_name, u32 value_default)
{
	DWORD D;
	char buffer[20] = {0};

	value = value_default;
	D = GetPrivateProfileString (section_name, value_name, 0, buffer, sizeof(buffer), PLUGIN_INI_NAME);
	if (D)
//	if (buffer[0])
	{
// crappy function return result; have to do a check twice
		if (D < sizeof(buffer) - 1)
		{
			A(buffer[0]);
			value = atoi (buffer);
		}
	}
}

// *requires Total Commander >= 5.51
void __stdcall PackSetDefaultParams (PackDefaultParamStruct* dps)
{
//	DWORD D;

	ReadConfigStringToU32 (packtype_hog, PLUGIN_SECTION_NAME, "hog", 0);
	ReadConfigStringToU32 (packtype_wad, PLUGIN_SECTION_NAME, "wad", 1);
}

// *not used by "Synchronize Dirs...", even by Total Commander 8.01
#if 0
int __stdcall StartMemPack (int options, char* filepath)
{
	ET;

	return 0;
}

int __stdcall PackToMem (int h_mem_pack, char* buf_in, int len_in, int* taken, char* buf_out, int len_out, int* written, int seek_by)
{
	ET;

	return 0;
}

int __stdcall DoneMemPack (int h_mem_pack)
{
	ET;

	return 0;
}
#endif

template<typename T>
bool TPackFiles_IsFile (u32& len, T* filename)
{
//	T c_last;
	u32 c_last;

	len = 0;
	while (*filename)
//	while ((c = *filename))
	{
		c_last = *filename;
		filename++;
		len++;
	}
	if (c_last == '\\')
		return 0;
	return 1;
}

struct pack_type_t
{
	const char* extension;
	const char* type;
};

// package types which we definitely know, of what format it is
pack_type_t pack_types[] =
{
	{"mn3", "hog2"},
	{"d3m", "hog2"},
	{"pak", "pak"},
//	{"", ""},
};

//bool PackFiles_SetPackFormat (int& result, PackTop* pack)
//bool PackFiles_GetPackFormatByFileExtension (int& result, const char*& format)
bool PackFiles_GetPackFormatByFileExtension (int& result, const char*& format, const char* extension)
{
//	bool r;
//	const char* type;

	format = 0;

// definite solutions
	FORI0(NUMOF(pack_types))
	{
		if (! _stricmp (extension, pack_types[i].extension))
		{
			format = pack_types[i].type;
			goto succeed;
		}
	}

// vague solutions
	if (! _stricmp (extension, "hog"))
	{
		switch (packtype_hog)
		{
		case 0:
			format = "hog";
			break;
		case 1:
			format = "hog2";
			break;
		default:
			ET;
		}
		goto succeed;
	}
	if (! _stricmp (extension, "wad"))
	{
		switch (packtype_hog)
		{
#if 0
		case 0:
			format = "wad";
			break;
#endif
		case 1:
			format = "wad2";
			break;
		case 2:
			format = "wad3";
			break;
		default:
			ET;
		}
		goto succeed;
	}

// we don't know what format to choose for this extension
	result = E_UNKNOWN_FORMAT;
//	result = E_NOT_SUPPORTED;
	return 0;

#if 0
set_handler:
	r = pack->SetHandlerByName (type);
//	if (! r)
//		goto pack_error;
	A(r);

	A(! pack->m_error);
	result = 0;
	return 1;
#else
succeed:
	result = 0;
	return 1;
#endif

#if 0
pack_error:
	result = ConvertError (pack);
	A(result);
	return 0;
#endif
}

template<typename T>
//int TPackFiles_OutputOpened (PackTop* pack, bool widechar, T* add_list)
//int TPackFiles_OutputOpened (PackTop* pack, bool widechar, T* src_path, T* add_list)
//bool TPackFiles_OutputOpened (PackTop* pack, bool widechar, T* src_path, T* add_list)
bool TPackFiles_OutputOpened (int& result, PackTop* pack, bool widechar, T* src_path, T* add_list)
{
	bool r;
//	u32 i;
	u32 files_num;
	T* ptr;
//	u32 bytes_processed;
	u32 len;

#if 0
//	r = pack->SetHandlerByName ("hog");
//	r = pack->SetHandlerByName ("hog2");
//	r = pack->SetHandlerByName ("pak");
//	r = pack->SetHandlerByName ("wad");
//	r = pack->SetHandlerByName ("wad3");
	r = PackFiles_SetPackFormat (result, pack);
	if (! r)
		return 0;
#endif

// count the number of files
// (this is so improper; should be given the number of files and array of strings instead)
	files_num = 0;
	ptr = add_list;
	while (*ptr)
	{
#if 0
		files_num++;
		do
			ptr++;
		while (*ptr);
#else
		r = TPackFiles_IsFile (len, ptr);
		files_num += r;
		ptr += len;
#endif
		ptr++;
	}
	r = pack->PrepareHdr (files_num);
	if (! r)
		goto pack_error;

//	bytes_processed = pack->m_data_begin;
	while (*add_list)
	{
		r = TPackFiles_IsFile (len, add_list);
		if (! r)
// skip this file
		{
			add_list += len;
			add_list++;
			continue;
		}

		pack->m_info_filename_is_widechar = widechar;
		pack->m_info_filename = add_list;
		r = ShowProgress (result, pack);
		if (! r)
			return 0;

		r = pack->CopyFromFile (widechar, src_path, add_list);
		if (! r)
		{
			goto pack_error;
		}
// skip the name of this file
//		add_list += len;
		do
			add_list++;
		while (*add_list);
		add_list++;
	}

	r = pack->FinishWrite ();
	if (! r)
		goto pack_error;

	A(! pack->m_error);
	result = 0;
	return 1;

pack_error:
	result = ConvertError (pack);
	A(result);
	return 0;
}

// not so convenient
#if 0
template<typename T>
bool TPackFiles_RemoveDirectoresFromList (int& result, T* list)
{
	char c_last;

	c_last = 0;
}
#endif

template<typename T>
//bool TPackFiles_GetExtension (int& result, PackTop* pack, T* filepath)
bool TPackFiles_GetExtension (int& result, char* buffer, u32 buffer_size, T* filepath)
{
	T* last_dot;
	T* last_backslash;
	T* extension;
//	T c;
	u32 c;
	u32 i;

	A(buffer_size);
	buffer[0] = 0;
	last_dot = 0;
	last_backslash = 0;
	while ((c = *filepath))
	{
		switch (c)
		{
		case '.':
			last_dot = filepath;
// signal, that the dot is going after a backslash
			last_backslash = 0;
			break;
		case '\\':
			last_backslash = filepath;
// if got a dot before, that will not be the extension
			last_dot = 0;
			break;
		}
		filepath++;
	}
	if (! last_dot)
		return 1;
	if (last_backslash)
		return 1;
	extension = last_dot + 1;
	for (i = 0; i < buffer_size; i++)
	{
// *note that it makes 0xffffff80 from 0x80
		c = extension[i];
		buffer[i] = c;
		if (! c)
			return 1;
		if (c < 0x20)
			goto not_supported;
		if (c >= 0x80)
			goto not_supported;
	}

not_supported:
	result = E_NOT_SUPPORTED;
	return 0;
}

// *it also receives directories :S
template<typename T>
//bool TPackFiles_ToPack (int& result, PackTop* pack, T* packed_file, T* src_path, T* add_list)
bool TPackFiles_ToPack (int& result, const char* format, PackTop* pack, T* packed_file, T* src_path, T* add_list)
{
	bool r;
	bool widechar;
	char file_ext[16];

// static condition
	widechar = sizeof(T) == 2;

	if (! format)
// we didn't get the target format; have to define it by ourselves by the extension of the file
	{
// *I don't see any other way to get the type of the package;
// maybe it is possible through some extension function ?
// no, the tatalcmd plugin port doesn't give any such function
//		r = TPackFiles_GetExtension (result, pack, packed_file);
		r = TPackFiles_GetExtension (result, file_ext, sizeof(file_ext), packed_file);
		if (! r)
			return 0;

#if 0
// *yet this is faily :  .wad will mean WAD 1, which we do not support, while we support WAD 2/3, which also have .wad extension
// same with HOG2, but you may compress to, lets say, .mn3, first, and then change the extension (rename the file)
//		r = pack->SetHandlerByName (pack->m_target_file_ext);
		r = pack->SetHandlerByName (file_ext);
		if (! r)
			goto pack_error;
// or this (the same)
//		format = file_ext;
#endif

//		format = file_ext;
//		r = PackFiles_GetPackFormatByFileExtension (result, format);
		r = PackFiles_GetPackFormatByFileExtension (result, format, file_ext);
		if (! r)
			return 0;
		A(format);
	}

	r = pack->SetHandlerByName (format);
	if (! r)
		goto pack_error;
//	A(r);

	r = pack->Open (1, widechar, packed_file);
	if (! r)
		goto pack_error;

//	r = TPackFiles_RemoveDirectoresFromList (result, add_list);

//	r = TPackFiles_OutputOpened (pack, widechar, src_path, add_list);
	r = TPackFiles_OutputOpened (result, pack, widechar, src_path, add_list);
	if (! r)
	{
		u32 err;
// backup the error
		err = pack->m_error;
		r = pack->Delete ();
		r = pack->Close ();
		pack->m_error = err;
		return 0;
	}

	r = pack->Close ();
	if (! r)
		goto pack_error;

	A(! pack->m_error);
	result = 0;
	return 1;

pack_error:
	result = ConvertError (pack);
	A(result);
	return 0;
}

// for different purposes/plugins we have two kinds of return results
template<typename T>
bool TPackFiles2 (int& result, u32& pack_result, const char* format, T* packed_file, T* src_path, T* add_list)
{
	bool r;
//	bool result;
//	int result;
	PackTop* pack;
//	u32 i;

	A(add_list);
	result = 0;
	pack_result = 0;

//	new0 (pack);
	pack = new PackTop;
	if (! pack)
	{
		result =  E_NO_MEMORY;
		pack_result = PACK_ERR__OUT_OF_MEMORY;
		goto clear_variables;
	}

	r = TPackFiles_ToPack (result, format, pack, packed_file, src_path, add_list);
	if (! r)
	{
		A(pack->m_error);
//		result = ConvertError (pack);
		pack_result = pack->m_error;
		goto finish;
	}

	A(! pack->m_error);

finish:
	delete pack;

clear_variables:
// *'SetProcessDataProc()' was called before 'PackFiles()'
	g_process_data_proc_w = 0;
	g_process_data_proc = 0;

//	return ! result;
	return ! pack_result;
}

template<typename T>
int TPackFiles (T* packed_file, T* sub_path, T* src_path, T* add_list, int flags)
{
	bool r;
//	bool result;
	int result;
//	u32 i;
// dummy
	u32 pack_result;

	A(add_list);
	if (sub_path)
		return E_NOT_SUPPORTED;
	if (flags & PK_PACK_MOVE_FILES)
		return E_NOT_SUPPORTED;
	if (flags & PK_PACK_ENCRYPT)
		return E_NOT_SUPPORTED;
//	result = 0;

	r = TPackFiles2 (result, pack_result, 0, packed_file, src_path, add_list);

	return result;
}

// *it isn't called in the case of deleting a file
// but is called in the case of adding a file to an existing pack, even though PK_CAPS_MODIFY is off
int __stdcall PackFiles (LPSTR packed_file, LPSTR sub_path, LPSTR src_path, LPSTR add_list, int flags)
{
	return TPackFiles (packed_file, sub_path, src_path, add_list, flags);
}

int __stdcall PackFilesW (wchar_t* packed_file, wchar_t* sub_path, wchar_t* src_path, wchar_t* add_list, int flags)
{
	return TPackFiles (packed_file, sub_path, src_path, add_list, flags);
}

// not planned to be supported
//int __stdcall DeleteFiles (LPSTR packed_file, LPSTR delete_list)
//int __stdcall DeleteFilesW (wchar_t* packed_file, wchar_t* delete_list)

// *remove me (we don't manage/enhance Totalcmd API)
#if 0
// ==== non WCX subroutines ====

bool __stdcall PackFilesToFile (u32& pack_result, const char* format, const char* packed_file, const char* src_path, const char* add_list)
{
	bool r;
// dummy
	int result;

//	__asm int 3
	r = TPackFiles2 (result, pack_result, format, packed_file, src_path, add_list);

	return r;
}
#endif

