// by Alexander Ilyin
// ID Games WAD files
// look Doom src

#include "wad.h"
#include "error.h"
#include "pack_top.h"
#include "macros.h"


#if 0
Wad::Wad ()
{
	memset (this, 0, sizeof(*this));
}
#endif

bool Wad::SetHdrBuf ()
{
	m_hdr_buf = &m_buf_hdr;
	m_hdr_size = sizeof(m_buf_hdr);
	return 1;
}

bool Wad::HdrParse (bool& is_valid)
{
	m_files_num = m_buf_hdr.numlumps;
	m_entries_begin = m_buf_hdr.infotableofs;
// some checks
	is_valid = m_entries_begin >= 4 + sizeof(m_buf_hdr);
	is_valid &= (m_entries_begin <= m_file.m_len);

	return 1;
}

bool Wad::TypeSetup ()
{
	bool r;

	r = SetupContainer ();
	if (! r)
		return 0;
	m_entry_data = &m_buf_entry;
	m_entry_len = sizeof(m_buf_entry);
	m_read_swap_filepointer = 1;
	m_data_offset_32 = &m_buf_entry.filepos;
//	m_name = m_buf_entry.name;
	m_name = m_name_buf;
// *in the intermediate buffer it will be 0 ended
	m_name_always_zero_ended = 1;
//	m_name_always_zero_ended = 0;
//	m_name_len_max = sizeof(m_buf_entry.name) - 1;
//	m_name_len_max = sizeof(m_buf_entry.name);
	m_name_len_max = sizeof(m_name_buf) - 1;
	m_data_len_32 = &m_buf_entry.filelen;
	m_entry_get_name = &Wad::EntryGetName;
//	m_entry_reader_post = &Wad::EntryReaderPost;
	m_entry_form = &Wad::EntryForm;
	m_write_finish = &Wad::WriteFinish;
	m_has_files_num = 1;
// + writer-only configuration +
	m_write_entries_after_data = 1;
	m_write_entries_store_in_a_buf = 1;
// 0 for Doom 1 and 2
	m_write_default_bytes_entry = 0;
	m_write_are_subdirectories_allowed = 0;
// - writer-only configuration -

// cleanup some our values
	m_type_cur = 0;

	return 1;
}

char* wad_markers[][2] =
{
	{0, 0},
// *"linuxdoom-1.10" has only these two
	{"F_START", "F_END"},
	{"S_START", "S_END"},
// user WAD files
	{"SS_START", "SS_END"},
	{"P_START", "P_END"},
// appearing in official WADs, but are unused by known Doom engines
	{"P1_START", "P1_END"},
	{"P2_START", "P2_END"},
	{"P3_START", "P3_END"},
	{"F1_START", "F1_END"},
	{"F2_START", "F2_END"},
};

bool Wad::EntryGetName ()
{
// copy to buffer
// check that the buffer would have enough place to add an extension
//	A(sizeof(m_name_buf) > sizeof(m_buf_entry.name));
//	A(sizeof(m_name_buf) > sizeof(m_buf_entry.name) + 1 + 3);
	A(sizeof(m_name_buf) > sizeof(m_buf_entry.name) + 1 + 5);
	memcpy (m_name_buf, m_buf_entry.name, sizeof(m_buf_entry.name));
// ensure 0 ending
	m_name_buf[sizeof(m_buf_entry.name)] = 0;

	FORI1(NUMOF(wad_markers))
	{
		if (! strcmp (m_name_buf, wad_markers[i][0]))
		{
//			m_type_cur = i;
// lets do a simple stacking (of only 1 value)
			if (! m_type_cur)
				m_type_cur = i;
			goto special_entry;
		}
		else if (! strcmp (m_name_buf, wad_markers[i][1]))
		{
// *should not do this, but stack the types ("linuxdoom-1.10" does a search, for the positions of these markers in the directory (look in "r_data.c"))
//			m_type_cur = 0;
// lets do a simple stacking (of only 1 value)
			if (m_type_cur == i)
				m_type_cur = 0;
			goto special_entry;
		}
	}

// apply extensions
	switch (m_type_cur)
	{
// does not belong to a group
	case 0:
		break;
	case 1:
// *idk the actual extension
		strcat (m_name_buf, ".flat");
		break;
	case 2:
	case 3:
// *idk the actual extension
		strcat (m_name_buf, ".spr");
		break;
	case 4:
// *idk the actual extension
		strcat (m_name_buf, ".patch");
		break;
#if 0
	default:
		m_error = PACK_ERR__UNSUPPORTED_ENTRY_TYPE;
		return 0;
#endif
	}

	return 1;

special_entry:
	m_error = PACK_ERR__SPECIAL_ENTRY;
	return 0;
}

#if 1
bool Wad::EntryReaderPost (bool& is_valid)
{
	u32 i;

// convert "/" to "\", otherwise Total Commander will fail to open such file
	for (i = 0; i < m_name_len_max; i++)
	{
		char c;
		c = m_name[i];
		if (! c)
			break;
		if (c == '/')
			m_name[i] = '\\';
	}
	return 1;
}
#endif

bool Wad::EntryForm ()
{
//	bool r;

	m_error = PACK_ERR__UNSUPPORTED_FORMAT_VERSION;
	return 0;
	m_buf_entry.filepos = m_data_offset_cur;
	m_buf_entry.filelen = m_data_len;
	return 1;
}

bool Wad::WriteFinish ()
{
	m_buf_hdr.numlumps = m_file_ix;
	m_buf_hdr.infotableofs = m_entries_begin;
	return 1;
}

